;*********************************************************************
; 4xD - Quad Digital Delay - DSP-PIC                                 *
;*********************************************************************
; (c) 2006 SM-Sound / Seb Francis                                    *
;*********************************************************************
;
; Main code
;
;*********************************************************************
;                                                               
;    Connections:
; 
;    RB0  ->  SRAM A1
;    RB1  ->  SRAM A2
;    RB2  ->  SRAM A3
;    RB3  ->  SRAM A4
;    RB4  ->  SRAM A5
;    RB5  ->  SRAM A6
;    RB6  <-> PGC & SRAM A7
;    RB7  <-> PGD & SRAM A8
;    RB8  ->  SRAM A9
;    RB9  ->  SRAM A10
;    RB10 ->  SRAM A11
;    RB11 ->  SRAM A12
;    RB12 ->  SRAM A13
;    RB13 ->  SRAM A14
;    RB14 ->  SRAM A15 / Optional Analog Input
;    RB15 ->  SRAM A16 / Optional Analog Input
;        
;    RC1  ->  Optional SRAM A15
;    RC2  ->  Optional SRAM A16
;    RC13 <-> EMUD1
;    RC14 <-> EMUC1
;    RC15 ->  SRAM A0
; 
;    RD0  <-> SRAM I/O0
;    RD1  <-> SRAM I/O1
;    RD2  <-> SRAM I/O2
;    RD3  <-> SRAM I/O3
;    RD4  <-> SRAM I/O4
;    RD5  <-> SRAM I/O5
;    RD6  <-> SRAM I/O6
;    RD7  <-> SRAM I/O7
;    RD8  ->  SRAM !OE
;    RD9  ->  SRAM !WE
;    RD10 ->  CLIP LED LEFT
;    RD11 ->  CLIP LED RIGHT
; 
;    RF0  ->  
;    RF1  ->  
;    RF2  <-  SPI SDI1
;    RF3  ->  SPI SDO1 (not active)
;    RF4  ->  
;    RF5  ->  
;    RF6  <-  SPI SCK1
; 
;    RG0  -> 
;    RG1  -> 
;    RG2  -> 
;    RG3  -> 
;    RG6  ->  Optional Digital I/O
;    RG7  ->  Optional Digital I/O
;    RG8  ->  Optional Digital I/O
;    RG9  ->  Optional Digital I/O
;    RG12 <-  CODEC CSDI
;    RG13 ->  CODEC CSDO
;    RG14 <-  CODEC CSCK
;    RG15 <-  CODEC COFS
; 
;===================================================================================

	.equ	__30F6012A, 1		; Define processor
	.include "p30f6012A.dsinc"
	
;-----------------------------------------------------------------------------------
; Global Declarations

	.global	__reset			; The label for the first line of code
	.global	__T1Interrupt		; Timer 1 interrupt
	.global	__T4Interrupt		; Timer 4 interrupt
	.global	__T5Interrupt		; Timer 5 interrupt
	.global	__SPI1Interrupt		; SPI interrupt
	.global	__DCIInterrupt		; DCI (CODEC) interrupt
	
	; Error trap globals are defined in file "traps.s"

;-----------------------------------------------------------------------------------
; Configuration bits

	config	__FOSC, CSW_FSCM_OFF & FRC_PLL16			; Clock switch off, Clock monitor off, PLLx16
	config	__FWDT, WDT_ON & WDTPSA_8 & WDTPSB_2			; Watchdog timer on, Prescale x16 = 32ms timeout
	config	__FBORPOR, PBOR_OFF & BORV_27 & PWRT_64 & MCLR_EN	; Brown out detect off, 64ms startup delay, MCLR on
	config	__FGS, CODE_PROT_OFF					; Code protect off
	
;-----------------------------------------------------------------------------------
; Constants

	; Defines which dsPIC this is (0 or 1)
	.equ	WHICH_DSPIC, #0
	
	; How close the audio levels need to be to switch delay time in non-resampling mode
	; Having this set too low makes the delay time change in a bit of a jumpy way (especially when controlled by fast LFO)
	; Having this set too high makes delay time changes a bit clicky
	; 0x0080 seems to be a nice setting
	.equ	LEVEL_MATCH_THRESHOLD, #0x0080
	
	; Maximum amounts to resample up/down in resampling mode
	; Buffer step:         0x0008 0x0010 0x0020 0x0040 0x0080 0x0100 0x0200 0x0400 0x0800 0x1000 0x2000
	; Resample by octaves:   -5     -4     -3     -2     -1      0     +1     +2     +3     +4     +5
	.equ	MIN_BUFFER_STEP, #0x0008	; Minimum output buffer step -5 octaves
	.equ	MAX_BUFFER_STEP, #0x2000	; Maximum output buffer step +5 octaves
	
	; If this is set to 1 then the audio is muted during start-up until the UI PIC has sent the current
	; parameters via SPI followed by an Enable Audio message.  For debugging it is recommended to set
	; this to 0 so the dsPICs can be restarted while the UI PIC is still running.
	.equ	MUTE_AUDIO_DURING_STARTUP, #0
	

	.include "const.dsinc"		; Constants + Memory map
	.include "sinetable.dsinc"	; Sine lookup table
	.include "fir.dsinc"		; FIR coefficients table and RAM definition
	
;===================================================================================
	.text				; Start of Code section
;-----------------------------------------------------------------------------------
; Initialize stack pointer and limit register

__reset:
	mov	#__SP_init, W15		; Initialize the Stack Pointer register
	mov	#__SPLIM_init, W0	; Get address at the end of stack space
	mov	W0, SPLIM		; Load the Stack Pointer Limit register
	nop				; Add NOP to follow SPLIM initialization
	
;-----------------------------------------------------------------------------------
; Initialisation

Init:
	disi	#16383			; Disable interrupts during initialisation

	; Initialise W registers
	clr	W0
	clr	W14
	repeat	#12
	mov	W0, [++W14]
	clr	W14
	
;	mov	#0b1100, W0		; Tune internal oscillator -3% so as
;	mov	W0, OSCTUN		; not to exceed SRAM timing specs
	
	; Configure I/O ports (all inputs by default)
	mov	#0xFFFF, W0		; Configure all A/D pins for digital I/O
	mov	W0, ADPCFG
	clr	TRISB			; RB* all outputs
	clr	TRISC			; RC* all outputs
	mov	#0b0000000011111111, W0	; RD0-7 inputs, RD8-11 outputs
	mov	W0, TRISD
	mov	#0b0000000001000100, W0	; RF* all outputs, except RF2, RF6
	mov	W0, TRISF
	mov	#0b1101000000000000, W0	; RG* all outputs, except RG12, RG14, RG15
	mov	W0, TRISG
	
	; Set up interrupts
;	bclr	INTCON1, #NSTDIS	; Enable nesting interrupts (default)
	bset	IPC10, #DCIIP0		; Raise priority of DCI interrupt to 5
	bclr	IFS0, #T1IF		; Clear the Timer1 interrupt flag
	bclr	IFS1, #T4IF		; Clear the Timer4 interrupt flag
	bclr	IFS1, #T5IF		; Clear the Timer5 interrupt flag
	bclr	IFS0, #SPI1IF		; Clear the SPI1 interrupt flag
	bclr	IFS2, #DCIIF		; Clear the DCI interrupt flag
	bset	IEC0, #T1IE		; Enable Timer1 interrupts
	bset	IEC1, #T4IE		; Enable Timer4 interrupts
	bset	IEC1, #T5IE		; Enable Timer4 interrupts
	bset	IEC0, #SPI1IE		; Enable SPI interrupts
					; Enable DCI interrupts after rest of initialisation
	; Don't need to change INTCON2 or IPCx
	
	; Set up Timer1
	mov	#0xFFFF, W0		; Load PR1 with 0xFFFF (slowest interval - interrupt every FCLK/65536 = 436 Hz @ 28.5956 MIPS = 2.3ms)
	mov	W0, PR1
	bset	T1CON, #TON		; Timer1 running (pre-scale defaults to 1:1)
					
	; Set up Timer4
	mov	#48828, W0		; Load PR4 with 48828 (interrupt after 0.25s)
	mov	W0, PR4
	clr	TMR4			; Zero Timer4
	bset	T4CON, #TCKPS1		; Timer4 pre-scale 1:64
					; Timer4 stopped (default)
					; Timer4&5 in independent 16bit mode (default)
					; Timer4 source internal clock FCY (default)
					
	; Set up Timer5
	mov	#48828, W0		; Load PR5 with 48828 (interrupt after 0.25s)
	mov	W0, PR5
	clr	TMR5			; Zero Timer5
	bset	T5CON, #TCKPS1		; Timer5 pre-scale 1:64
					; Timer5 stopped (default)
					; Timer5 source internal clock FCY (default)
	
	; Set up SPI
	bset	SPI1STAT, #SPIEN	; Enable SPI
	bset	SPI1CON, #DISSDO	; SDO1 pin not used
	bset	SPI1CON, #MODE16	; 16 bit wide commication
					; Framed SPI support disabled (default)
					; Data on transition from idle to active clock state (default)
					; SS pin not used (default)
					; Clock idle state low (default)
					; Slave mode (default)
	
	; Set up DCI
	clr	TXBUF0			; Zero DCI output buffers
	clr	TXBUF1
	clr	TXBUF2
	clr	TXBUF3
	bset	DCICON1, #DCIEN		; DCI enabled
	bset	DCICON1, #CSCKD		; CSCK (bit clock) pin is input
	bset	DCICON1, #CSCKE		; Data changes on falling clock edge, sampled on rising
	bset	DCICON1, #COFSD		; COFS (frame sync) pin is input
					; Data transmission/reception is begun one serial clock cycle after frame synchronization pulse (default)
	bset	DCICON1, #COFSM0	; I2S frame sync mode
	bset	DCICON2, #BLEN1		; 4 data words (64 bits) buffered between interrupts
	bset	DCICON2, #BLEN0
	bset	DCICON2, #COFSG1	; Data frame has 4 words (64 bits)
	bset	DCICON2, #COFSG0
	bset	DCICON2, #WS3		; Data word size is 16 bits
	bset	DCICON2, #WS2
	bset	DCICON2, #WS1
	bset	DCICON2, #WS0
	; Don't need to change DCICON3 or DCISTAT
	mov	#0xFFFF, W0
	mov	W0, RSCON		; Enable receive for all time slots
	mov	W0, TSCON		; Enable transmit for all time slots
	
	; Map 0x8000+ data memory to program memory
	bset	CORCON, #PSV		; Enable program space visibility
	mov	#psvpage(SineTable), W0	; Point at sine table
	mov	W0, PSVPAG
	
	; Initialise output pins
	bclr	LATD, #10		; Clip LEDs off
	bclr	LATD, #11
	bclr	LATD, #8		; SRAM Output Enable
	bset	LATD, #9		; SRAM Write Disable (Read mode)
	bset	LATC, #15		; SRAM A0 = 1
	
	; DSP engine setup
;	bclr	CORCON, #US		; Signed DSP multiply instructions (default)
	bset	CORCON, #SATA		; Accumulator A saturation enabled
	bset	CORCON, #SATB		; Accumulator B saturation enabled
	bset	CORCON, #SATDW		; Data space write saturation enabled
;	bclr	CORCON, #ACCSAT		; Accumulator saturation mode 1.31 (default)
;	bclr	CORCON, #RND		; 'Unbiased' accumulator round mode (default)
	bset	CORCON, #IF		; Integer operation for DSP instructions
	
	; FIR setup
	; Initialise coefficients from program space
	mov	#tbloffset(FIR_CoeffsTable), W0
	mov	#FIR_CoeffsX, W1
	repeat	#SAMPLES-1
	tblrdl	[W0++], [W1++]
	mov	#tbloffset(FIR_CoeffsTable), W0
	mov	#FIR_CoeffsY, W1
	repeat	#SAMPLES-1
	tblrdl	[W0++], [W1++]
	; Clear FIR buffers
	mov	#FIR_BufferL, W0
	repeat	#SAMPLES-1
	clr	[W0++]
	mov	#FIR_BufferR, W0
	repeat	#SAMPLES-1
	clr	[W0++]
	clr	FIR_ResultL
	clr	FIR_ResultR
	
	; Modulus addressing setup
	mov	#0b1100000010101000, W0	; X AGU modulus addressing enabled
	mov	W0, MODCON		; Y AGU modulus addressing enabled
					; XWM = WXS (W8), YWM = WYS (W10)
	mov	#FIR_BufferL_End, WXS
	mov	WXS, XMODEND
	mov	#FIR_BufferL, WXS
	mov	WXS, XMODSRT
	mov	#FIR_BufferR_End, WYS
	mov	WYS, YMODEND
	mov	#FIR_BufferR, WYS
	mov	WYS, YMODSRT

	; Initialise variables
	clr	BufferL_InOffset
	clr	BufferL_OutOffset
	clr	BufferL_OutOffset_L
	clr	BufferL_OutOffset_New
	clr	BufferR_InOffset
	clr	BufferR_OutOffset
	clr	BufferR_OutOffset_L
	clr	BufferR_OutOffset_New
	mov	#0x100, W0
	mov	W0, BufferL_OutStep
	mov	W0, BufferR_OutStep
	
	clr	DelayTimeL
	clr	DelayTimeR
	clr	BaseDelayTimeL
	clr	BaseDelayTimeR
	clr	LFO_Rate
	clr	LFO_Step
	clr	LFO_AmpOffset
	clr	LFO_PhaseOffset
	clr	LFO_AngleL
	clr	LFO_AngleL_L
	clr	LFO_AngleROffset
	clr	LFO_DepthL
	clr	LFO_DepthR
	clr	LFO_Wave
	clr	LFO_Sync
	clr	LFO_SyncClock
	clr	Flags
	clr	DryMixL
	clr	DryMixR
	clr	DryMixFactorL
	clr	DryMixFactorR
	mov	#0xFFFF, W0
	mov	W0, WetMixFactorL
	mov	W0, WetMixFactorR
	clr	FeedbackFactorL
	clr	FeedbackFactorR
	
	; Initialise useful numbers
	mov	#49, W0
	mov	W0, N49
	mov	#74, W0
	mov	W0, N74
	mov	#114, W0
	mov	W0, N114
	mov	#189, W0
	mov	W0, N189
	mov	#200, W0
	mov	W0, N200
	mov	#219, W0
	mov	W0, N219
	mov	#6695, W0
	mov	W0, N6695
	mov	#32768, W0
	mov	W0, N32768
	mov	#MIN_BUFFER_STEP, W0
	mov	W0, N_MIN_BUFFER_STEP
	mov	#MAX_BUFFER_STEP, W0
	mov	W0, N_MAX_BUFFER_STEP
	mov	#0x100, W0
	mov	W0, N0x100
	
	clr	DISICNT			; Re-enable interrupts
	
	; Clear SRAM
	clr	W_SRAMData
	clr	W_SRAMAddress
	mov	#65535,W0
	do	W0, DoEnd1
	call	WriteSRAM
DoEnd1:	inc	W_SRAMAddress, W_SRAMAddress


	bset	IEC2, #DCIIE		; Enable DCI interrupts
	
	
	; Flash LEDs to indicate starting up
	; This is for debugging purposes to see if the code crashes
	; sometimes and gets reset by the watchdog timer.
	bset	LATD, #10		; Turn on clip LED 1
	bset	T4CON, #TON		; Start Timer4
	btsc	LATD, #10		; Loop while clip LED 1 is on
	bra	$-2
	bset	LATD, #11		; Turn on clip LED 2
	bset	T5CON, #TON		; Start Timer5
	btsc	LATD, #11		; Loop while clip LED 2 is on
	bra	$-2
	bset	LATD, #10		; Turn on clip LED 1
	bset	T4CON, #TON		; Start Timer4
	btsc	LATD, #10		; Loop while clip LED 1 is on
	bra	$-2
	bset	LATD, #11		; Turn on clip LED 2
	bset	T5CON, #TON		; Start Timer5
	btsc	LATD, #11		; Loop while clip LED 2 is on
	bra	$-2
	bset	LATD, #10		; Turn on clip LED 1
	bset	T4CON, #TON		; Start Timer4
	btsc	LATD, #10		; Loop while clip LED 1 is on
	bra	$-2
	bset	LATD, #11		; Turn on clip LED 2
	bset	T5CON, #TON		; Start Timer5
	btsc	LATD, #11		; Loop while clip LED 2 is on
	bra	$-2
	bset	LATD, #10		; Turn on clip LED 1
	bset	T4CON, #TON		; Start Timer4
	btsc	LATD, #10		; Loop while clip LED 1 is on
	bra	$-2
	bset	LATD, #11		; Turn on clip LED 2
	bset	T5CON, #TON		; Start Timer5
	btsc	LATD, #11		; Loop while clip LED 2 is on
	bra	$-2
	bset	LATD, #10		; Turn on clip LED 1
	bset	T4CON, #TON		; Start Timer4
	btsc	LATD, #10		; Loop while clip LED 1 is on
	bra	$-2
	bset	LATD, #11		; Turn on clip LED 2
	bset	T5CON, #TON		; Start Timer5
	btsc	LATD, #11		; Loop while clip LED 2 is on
	bra	$-2
	bset	LATD, #10		; Turn on clip LED 1
	bset	T4CON, #TON		; Start Timer4
	btsc	LATD, #10		; Loop while clip LED 1 is on
	bra	$-2
	bset	LATD, #11		; Turn on clip LED 2
	bset	T5CON, #TON		; Start Timer5
	btsc	LATD, #11		; Loop while clip LED 2 is on
	bra	$-2
	bset	LATD, #10		; Turn on clip LED 1
	bset	T4CON, #TON		; Start Timer4
	btsc	LATD, #10		; Loop while clip LED 1 is on
	bra	$-2
	bset	LATD, #11		; Turn on clip LED 2
	bset	T5CON, #TON		; Start Timer5
	btsc	LATD, #11		; Loop while clip LED 2 is on
	bra	$-2
	bset	LATD, #10		; Turn on clip LED 1
	bset	T4CON, #TON		; Start Timer4
	btsc	LATD, #10		; Loop while clip LED 1 is on
	bra	$-2
	bset	LATD, #11		; Turn on clip LED 2
	bset	T5CON, #TON		; Start Timer5
	btsc	LATD, #11		; Loop while clip LED 2 is on
	bra	$-2
	

;-----------------------------------------------------------------------------------
; Main program

Main:
	btss	Flags, #DoFIR_L		; Need to do FIR calculation?
	bra	FIR_Done_L
	
	mov	FIR_SampleInL1, W0	; Save new samples to FIR buffer and point to oldest sample
	mov	W0, [WXS++]
	mov	FIR_SampleInL2, W0
	mov	W0, [WXS++]
	mov	#FIR_CoeffsY, WYC	; Set pointer to first coefficient in Y RAM table
	clr	A, [WXS]+=2,WXD, [WYC]+=2,WYD		; Zero accumulator and pre-fetch X & Y values
	repeat	#SAMPLES-2
	mac	WXD*WYD, A, [WXS]+=2,WXD, [WYC]+=2,WYD	; A += WXD*WYD, pre-fetch next WXD & WYD
	mac	WXD*WYD, A				; Do last MAC separately to preserve pointers
	sftac	A, #-1		; A *= 2
	sac.r	A, W0		; Round and copy result to FIR_ResultL
	mov	W0, FIR_ResultL
	
	bclr	Flags, #DoFIR_L		; Clear flag
	
	; Accumulator saturation occurred?
	btss	SR, #SA
	bra	FIR_Done_L
	bclr	SR, #SA
	bset	LATD, #10		; Turn on clip LED 1
	bset	T4CON, #TON		; Start Timer4
	
FIR_Done_L:
	btss	Flags, #DoFIR_R		; Need to do FIR calculation?
	bra	FIR_Done_R
	
	mov	FIR_SampleInR1, W0	; Save new samples to FIR buffer and point to oldest sample
	mov	W0, [WYS++]
	mov	#FIR_BufferR_End, W1
	cp	W1, WYS		; If WYS > FIR_BufferR_End, reset to FIR_BufferR (MCU instructions don't support Y modulo addressing)
	btsc	SR, #N
	mov	#FIR_BufferR, WYS
	mov	FIR_SampleInR2, W0
	mov	W0, [WYS++]
	cp	W1, WYS		; If WYS > FIR_BufferR_End, reset to FIR_BufferR (MCU instructions don't support Y modulo addressing)
	btsc	SR, #N
	mov	#FIR_BufferR, WYS
	mov	#FIR_CoeffsX, WXC	; Set pointer to first coefficient in X RAM table
	clr	A, [WXC]+=2,WXD, [WYS]+=2,WYD		; Zero accumulator and pre-fetch X & Y values
	repeat	#SAMPLES-2
	mac	WXD*WYD, A, [WXC]+=2,WXD, [WYS]+=2,WYD	; A += WXD*WYD, pre-fetch next WXD & WYD
	mac	WXD*WYD, A				; Do last MAC separately to preserve pointers
	sftac	A, #-1		; A *= 2
	sac.r	A, W0		; Round and copy result to FIR_ResultR
	mov	W0, FIR_ResultR
	
	bclr	Flags, #DoFIR_R		; Clear flag
	
	; Accumulator saturation occurred?
	btss	SR, #SA
	bra	FIR_Done_R
	bclr	SR, #SA
	bset	LATD, #11		; Turn on clip LED 2
	bset	T5CON, #TON		; Start Timer5

FIR_Done_R:
	inc2	IdleCounter
	inc2	IdleCounter
	bra	Main
	
	
;===================================================================================
; Subroutines

; Reads 16 bit word from SRAM
; 
; Call with 16 bit word address in W_SRAMAddress
; Returns with data in W_SRAMData
; Changes W0
ReadSRAM:
	mov	W_SRAMAddress, LATB	; Select memory address
	nop
	mov	PORTD, W_SRAMData	; Get MSbyte
	swap	W_SRAMData
	bclr	LATC, #15		; Point to LSbye
	nop
	mov	PORTD, W0		; Get LSbyte
	mov.b	W0, W_SRAMData
	bset	LATC, #15		; Point to MSbyte again
	return
	

; Writes 16 bit word to SRAM
;
; Call with 16 bit word address in W_SRAMAddress
;           data in W_SRAMData
; Changes W0
WriteSRAM:
	mov	W_SRAMAddress, LATB ; Select memory address
	bclr	LATD, #9		; SRAM Write Enable
	clr	TRISD			; RD* all outputs
	lsr	W_SRAMData, #8, W0	; MSbyte to W0L
	mov.b	WREG, LATD		; Set MSbyte
	nop
	bclr	LATC, #15		; Point to LSbye
	mov	W_SRAMData, W0		; Set LSbyte
	mov.b	WREG, PORTD
	bset	LATD, #8		; SRAM Output Disable
	bset	LATD, #9		; SRAM Write Disable
	mov	#0b0000000011111111, W0	; RD0-7 inputs, RD8-11 outputs
	mov	W0, TRISD
	bclr	LATD, #8		; SRAM Output Enable
	bset	LATC, #15		; Point to MSbyte again
	return
	
	
;===================================================================================
;===================================================================================

__T1Interrupt:
	clrwdt				; Reset watchdog timer
	bclr	IFS0, #T1IF		; Clear the Timer1 interrupt flag
	retfie				; Return from ISR
	
;===================================================================================

__T4Interrupt:
	bclr	LATD, #10		; Turn off clip LED 1
	bclr	T4CON, #TON		; Stop Timer4
	clr	TMR4			; Zero Timer4
	bclr	IFS1, #T4IF		; Clear the Timer4 interrupt flag
	retfie				; Return from ISR
	
;===================================================================================

__T5Interrupt:
	bclr	LATD, #11		; Turn off clip LED 2
	bclr	T5CON, #TON		; Stop Timer5
	clr	TMR5			; Zero Timer5
	bclr	IFS1, #T5IF		; Clear the Timer5 interrupt flag
	retfie				; Return from ISR
	
;===================================================================================

__SPI1Interrupt:
	push	SR			; Save SR flags and W0-W3 this way as shadow registers are used by high priority DCI interrupt
	push.d	W0			; Save W0, W1
	push.d	W2			; Save W2, W3
	
	mov	SPI1BUF, W0		; Read received data
	
.if (WHICH_DSPIC == 0)			; Determine if this dsPIC is being addressed
	btsc	W0, #15
.else
	btss	W0, #15
.endif
	bra	__SPI1Interrupt_End
	
	lsr	W0, #8, W1		; W0H to W1L
	and	#0b01111111, W1		; AND to leave lowest 7 bits of message type in W1
	and	#0b11111111, W0		; Leave message data only in W0
	
	; 0000 (0)  -> Delay Time L
	; 0001 (1)  -> Delay Time R
	; 0010 (2)  -> LFO Rate
	; 0011 (3)  -> LFO Depth
	; 0100 (4)  -> LFO amplitude offset / LFO Phase offset
	; 0101 (5)  -> LFO Waveform
	; 0110 (6)  -> LFO Sync Start - transmitted when MIDI Start message received, data byte is sync type
	; 0111 (7)  -> Resampling mode (off, on)
	; 1000 (8)  -> Digital dry mix level L
	; 1001 (9)  -> Digital dry mix level R
	; 1010 (10) -> Digital feedback level L
	; 1011 (11) -> Digital feedback level R
	; 1100 (12) -> LFO Sync Clock Period L
	; 1101 (13) -> LFO Sync Clock Period H
	; 1110 (14) -> Double delay time
	; 1111 (15) -> Enable Audio (off, on)
	
	cp	W1, #6
	bra	Z, SPI_LFOSyncStart
	cp	W1, #12
	bra	Z, SPI_LFOSyncClock_L
	cp	W1, #13
	bra	Z, SPI_LFOSyncClock_H
	cp	W1, #0
	bra	Z, SPI_DelayTimeL
	cp	W1, #1
	bra	Z, SPI_DelayTimeR
	cp	W1, #2
	bra	Z, SPI_LFORate
	cp	W1, #3
	bra	Z, SPI_LFODepth
	cp	W1, #4
	bra	Z, SPI_LFOAmpPhase
	cp	W1, #5
	bra	Z, SPI_LFOWaveform
	cp	W1, #7
	bra	Z, SPI_ResamplingMode
	cp	W1, #8
	bra	Z, SPI_DryMixL
	cp	W1, #9
	bra	Z, SPI_DryMixR
	cp	W1, #10
	bra	Z, SPI_FeedbackL
	cp	W1, #11
	bra	Z, SPI_FeedbackR
	cp	W1, #14
	bra	Z, SPI_DoubleDelayTime
	cp	W1, #15
	bra	Z, SPI_EnableAudio
	bra	__SPI1Interrupt_End
	

; ---------------------------------------------------------------------------
	
	
SPI_DelayTimeL:
	; Call with delay time from UI PIC in W0
	; Returns with actual delay time in ms/10 in W0
	call	ConvertDelayTimeToMs
	
	; Number of samples: (Samples per second) / 1000 * (10ths ms) / 10
	;                  = 48828.125 * (10ths ms) / 10,000
	;                  = 20000 * (10ths ms) / 4096
	mov	#20000, W1
	mul.uu	W0, W1, W0
	sl	W1, #4, W1
	lsr	W0, #12, W0
	ior	W0, W1, W0
	
	mov	W0, BaseDelayTimeL
	
	; Update LFO_Depth
	bra	SPI_LFODepth_ReCalc


; ---------------------------------------------------------------------------


SPI_DelayTimeR:
	; Call with delay time from UI PIC in W0
	; Returns with actual delay time in ms/10 in W0
	call	ConvertDelayTimeToMs
	
	; Number of samples: (Samples per second) / 1000 * (10ths ms) / 10
	;                  = 48828.125 * (10ths ms) / 10,000
	;                  = 20000 * (10ths ms) / 4096
	mov	#20000, W1
	mul.uu	W0, W1, W0
	sl	W1, #4, W1
	lsr	W0, #12, W0
	ior	W0, W1, W0
	
	mov	W0, BaseDelayTimeR
	
	; Update LFO_Depth
	bra	SPI_LFODepth_ReCalc
	

; ---------------------------------------------------------------------------


SPI_LFORate:
	mov	W0, LFO_Rate
	
	; LFO Rate * 64 => W0
	mov	#64, W1
	mul.uu	W0, W1, W0
	
	; 16382 - W0 => W0
	mov	#16382, W1
	sub	W1, W0, W0
	
	; Do sine table lookup
	mov	#psvoffset(SineTable), W1
	add	W0, W1, W0
	mov	[W0], W1
	
	; W1 = 0x7FFF - W1
	mov	#0x7FFF, W0
	sub	W0, W1, W1

	mov	W1, LFO_Step

	bra	__SPI1Interrupt_End


; ---------------------------------------------------------------------------


SPI_LFODepth:
	; Save value
	mov	W0, LFO_Depth
	
SPI_LFODepth_ReCalc:
	btsc	Flags, #SingleInput	; Single input mode?
	bra	SPI_LFODepth_SingleInput
	
	; If BaseDelayTime <= 0x3FFF
	; LFO_Depth = BaseDelayTime * LFO_Depth/200
	; If BaseDelayTime >= 0x4000
	; LFO_Depth = (0x7FFF - BaseDelayTime) * LFO_Depth/200
	mov	BaseDelayTimeL, W1
	mov	#0x7FFF, W2
	btsc	BaseDelayTimeL, #14
	sub	W2, W1, W1
	call	W1xLFO_Depthdiv200
	mov	W1, LFO_DepthL
	
	mov	BaseDelayTimeR, W1
	mov	#0x7FFF, W2
	btsc	BaseDelayTimeR, #14
	sub	W2, W1, W1
	call	W1xLFO_Depthdiv200
	
	bra	ScaleBy_LFO_AmpOffset
	
	
SPI_LFODepth_SingleInput:
	; If BaseDelayTime <= 0x7FFF
	; LFO_Depth = BaseDelayTime * LFO_Depth/200
	; If BaseDelayTime >= 0x8000
	; LFO_Depth = (0xFFFF - BaseDelayTime) * LFO_Depth/200
	mov	BaseDelayTimeL, W1
	mov	#0x7FFF, W2
	btsc	BaseDelayTimeL, #15
	sub	W2, W1, W1
	call	W1xLFO_Depthdiv200
	mov	W1, LFO_DepthL
	
	mov	BaseDelayTimeR, W1
	mov	#0x7FFF, W2
	btsc	BaseDelayTimeR, #15
	sub	W2, W1, W1
	call	W1xLFO_Depthdiv200
	
	
ScaleBy_LFO_AmpOffset:
	
	; Scale by LFO_AmpOffset
	; W1 * LFO_AmpOffset/20 -> LFO_DepthR
	
	; If LFO_AmpOffset = 20, ensure LFO_DepthR = W1
	mov	LFO_AmpOffset, W0
	cp	W0, #20
	bra	Z, LFO_AmpOffset_EQ20
	
	; W1 / 20 -> W1
	mov	#0xCCC, W2
	mul.uu	W1, W2, W0
	; W1 * LFO_AmpOffset -> LFO_DepthR
	mov	LFO_AmpOffset, W0
	mul.uu	W1, W0, W0
	mov	W0, LFO_DepthR
	
	bra	__SPI1Interrupt_End
	
LFO_AmpOffset_EQ20:
	mov	W1, LFO_DepthR
	
	bra	__SPI1Interrupt_End

; W1 = W1 * LFO_Depth/200
W1xLFO_Depthdiv200:
	mov	LFO_Depth, W0
	cp	N200
	btsc	SR, #Z
	return
	; W1 / 200 -> W3.W2
	mov	#0x148, W2
	mul.uu	W1, W2, W2
	; W3.W2 * LFO_Depth -> W1
	mul.uu	W3, W0, W0
	mov	LFO_Depth, W1
	mul.uu	W2, W1, W2
	add	W0, W3, W1
	
	return
	

; ---------------------------------------------------------------------------


SPI_LFOAmpPhase:
	; W0 < 20 ?
	cp	W0, #20
	bra	N, SPI_LFOAmp
	
; W0 >= 20
SPI_LFOPhase:
	; LFO_AmpOffset = 20 (0%)
	mov	#20, W1
	mov	W1, LFO_AmpOffset
	; LFO_DepthR = LFO_DepthL
	mov	LFO_DepthL, W1
	mov	W1, LFO_DepthR
	
	; W0 -= 20 * 2 -> LFO_PhaseOffset
	sub	W0, #20, W0
	sl	W0, #1, W0
	mov	W0, LFO_PhaseOffset
	
	; LFO_AngleROffset = LFO_AngleL + LFO_PhaseOffset * 182
	
	; LFO_PhaseOffset * 182 -> LFO_AngleROffset
	mov	#182, W1
	mul.uu	W0, W1, W0
	mov	W0, LFO_AngleROffset
	
	; Update LFO_Depth
	bra	SPI_LFODepth_ReCalc
	
	
; W0 < 20
SPI_LFOAmp:
	; LFO_PhaseOffset = 0 (0deg)
	clr	LFO_PhaseOffset
	; LFO_AngleROffset = 0
	clr	LFO_AngleROffset
	
	; W0 -> LFO_AmpOffset
	mov	W0, LFO_AmpOffset
	
	; LFO_DepthR = LFO_DepthL / 20 * LFO_AmpOffset
	
	; LFO_DepthL / 20 -> W1
	mov	LFO_DepthL, W0
	mov	#0xCCC, W1
	mul.uu	W0, W1, W0
	; W1 * LFO_AmpOffset -> LFO_DepthR
	mov	LFO_AmpOffset, W0
	mul.uu	W1, W0, W0
	mov	W0, LFO_DepthR

	; Update LFO_Depth
	bra	SPI_LFODepth_ReCalc


; ---------------------------------------------------------------------------


SPI_LFOWaveform:
	mov	W0, LFO_Wave
	
	bra	__SPI1Interrupt_End
	

; ---------------------------------------------------------------------------
	
	
SPI_LFOSyncStart:
	cp0	W0		; If LFO Sync = 0 (disabled), reset to last received LFO_Rate value
	btss	SR, #Z
	bra	SPI_LFOSyncStart_1
	mov	W0, LFO_Sync
	mov	LFO_Rate, W0
	bra	SPI_LFORate
	
SPI_LFOSyncStart_1:
	call	LookupLFOSync
	mov	W0, LFO_Sync
	
	; Zero LFO angle
	clr	LFO_AngleL
	
	bra	__SPI1Interrupt_End
	
	
; 1/48 bars per sync period
LookupLFOSync:
	bra	W0
	retlw	#0, W0		; 0=Off
	retlw	#768, W0	; 1=16
	retlw	#384, W0	; 2=8
	retlw	#192, W0	; 3=4
	retlw	#96, W0		; 4=2
	retlw	#48, W0		; 5=1
	retlw	#36, W0		; 6=3/4
	retlw	#24, W0		; 7=1/2
	retlw	#16, W0		; 8=1/3
	retlw	#12, W0		; 9=1/4
	retlw	#8, W0		; 10=1/6
	retlw	#6, W0		; 11=1/8
	retlw	#4, W0		; 12=1/12
	retlw	#3, W0		; 13=1/16
	

; ---------------------------------------------------------------------------


SPI_LFOSyncClock_L:
	mov	W0, LFO_SyncClock	; Save low byte of LFO_SyncClock
	
	bra	__SPI1Interrupt_End


; ---------------------------------------------------------------------------


SPI_LFOSyncClock_H:
	swap	W0
	ior	LFO_SyncClock		; Merge in high byte of LFO_SyncClock
	
	; If LFO sync disabled do nothing
	cp0	LFO_Sync
	bra	Z, __SPI1Interrupt_End
	
	; LFO_Step = 0x100,0000 / (LFO_Sync * LFO_SyncClock / 4)
	;          = 0x400,0000 / (LFO_Sync * LFO_SyncClock)
	;          = 0x4,0000 / (LFO_Sync * LFO_SyncClock / 0x100)
	
	; LFO_Sync * LFO_SyncClock => W3,W2
	mov	LFO_Sync, W2
	mov	LFO_SyncClock, W3
	mul.uu	W2, W3, W2
	
	; W3,W2 / 0x0100 => W2
	mov.b	W3, W2
	swap	W2
	
	; If won't fit into 16bits, set to max (0xFFFF)
	mov	#0xFF00, W0
	and	W3, W0, W0
	btss	SR, #Z
	mov	#0xFFFF, W2
	
	; 0x4,0000 / W2 => LFO_Step
	mov	#0x4, W1
	clr	W0
	repeat	#17
	div.ud	W0, W2
	mov	W0, LFO_Step
	
	bra	__SPI1Interrupt_End


; ---------------------------------------------------------------------------


SPI_ResamplingMode:
	btsc	W0, #0
	bra	SPI_ResamplingMode_On


SPI_ResamplingMode_Off:
	; Don't turn ResamplingMode if HalfSampleRate flag set
	btsc	Flags, #HalfSampleRate
	bra	__SPI1Interrupt_End

	disi	#((SPI_ResamplingMode_Off_1 - $)/2)	; Disable interrupts while messing with the buffers

	bclr	Flags, #ResamplingMode
	; BufferL_OutOffset_New = BufferL_InOffset - DelayTimeL
	mov	DelayTimeL, W0
	sub	BufferL_InOffset, WREG
	mov	W0, BufferL_OutOffset_New
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	BufferL_OutOffset_New, #15
	; BufferR_OutOffset_New = BufferR_InOffset - DelayTimeR
	mov	DelayTimeR, W0
	sub	BufferR_InOffset, WREG
	mov	W0, BufferR_OutOffset_New
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	BufferR_OutOffset_New, #15
	; BufferL_OutOffset_L, BufferR_OutOffset_L = 0
	clr	BufferL_OutOffset_L
	clr	BufferR_OutOffset_L
	; BufferL_OutStep, BufferR_OutStep = 0x100
	mov	0x100, W0
	mov	W0, BufferL_OutStep
	mov	W0, BufferR_OutStep
SPI_ResamplingMode_Off_1:
	bra	__SPI1Interrupt_End

	
SPI_ResamplingMode_On:
	bset	Flags, #ResamplingMode
	
	bra	__SPI1Interrupt_End


; ---------------------------------------------------------------------------


SPI_DryMixL:
	mov	W0, DryMixL
	
	; DryMixFactorL = 0x10000 * DryMixL       / 100
	; WetMixFactorL = 0x10000 * (100-DryMixL) / 100
	
	; 0x10000 * DryMixL => W1,W0
	mov	W0, W1
	clr	W0
	
	; 100 => W2
	mov	#100, W2
	
	; W1, W0 / W2 => DryMixFactorL
	repeat #17
	div.ud	W0, W2
	mov	W0, DryMixFactorL
	
	; 0x10000 * (100-DryMixL) => W1,W0
	mov	#100, W1
	mov	DryMixL, W0
	sub	W1, W0, W1
	clr	W0
	
	; W1, W0 / W2 => WetMixFactorL
	repeat #17
	div.ud	W0, W2
	; If overload (i.e. when DryMixL=0), set to 0xFFFF
	btsc	SR, #OV
	mov	#0xFFFF, W0
	mov	W0, WetMixFactorL
	
	bra	__SPI1Interrupt_End


; ---------------------------------------------------------------------------


SPI_DryMixR:
	mov	W0, DryMixR
	
	; DryMixFactorR = 0x10000 * DryMixR       / 100
	; WetMixFactorR = 0x10000 * (100-DryMixR) / 100
	
	; 0x10000 * DryMixR => W1,W0
	mov	W0, W1
	clr	W0
	
	; 100 => W2
	mov	#100, W2
	
	; W1, W0 / W2 => DryMixFactorR
	repeat #17
	div.ud	W0, W2
	mov	W0, DryMixFactorR
	
	; 0x10000 * (100-DryMixR) => W1,W0
	mov	#100, W1
	mov	DryMixR, W0
	sub	W1, W0, W1
	clr	W0
	
	; W1, W0 / W2 => WetMixFactorR
	repeat #17
	div.ud	W0, W2
	; If overload (i.e. when DryMixR=0), set to 0xFFFF
	btsc	SR, #OV
	mov	#0xFFFF, W0
	mov	W0, WetMixFactorR
	
	bra	__SPI1Interrupt_End
	
	
; ---------------------------------------------------------------------------


SPI_FeedbackL:
	; FeedbackFactorL = 0x10000 * FeedbackL / 100
	
	; 0x10000 * FeedbackL => W1,W0
	mov	W0, W1
	clr	W0
	
	; W1, W0 / 100 => FeedbackFactorL
	mov	#100, W2
	repeat #17
	div.ud	W0, W2
	mov	W0, FeedbackFactorL
	
	bra	__SPI1Interrupt_End
	
	
; ---------------------------------------------------------------------------


SPI_FeedbackR:
	; FeedbackFactorR = 0x10000 * FeedbackR / 100
	
	; 0x10000 * FeedbackR => W1,W0
	mov	W0, W1
	clr	W0
	
	; W1, W0 / 100 => FeedbackFactorR
	mov	#100, W2
	repeat #17
	div.ud	W0, W2
	mov	W0, FeedbackFactorR
	
	bra	__SPI1Interrupt_End
	
	
; ---------------------------------------------------------------------------


SPI_DoubleDelayTime:
	cp	W0, #2
	bra	Z, SPI_DoubleDelayTime_2
	cp	W0, #1
	bra	Z, SPI_DoubleDelayTime_1

SPI_DoubleDelayTime_0:
	bclr	Flags, #HalfSampleRate
	bclr	Flags, #SingleInput
	bclr	BufferL_InOffset, #15
	bclr	BufferL_OutOffset, #15
	bclr	BufferL_OutOffset_New, #15
	bclr	DelayTimeL, #15
	bclr	BufferR_InOffset, #15
	bclr	BufferR_OutOffset, #15
	bclr	BufferR_OutOffset_New, #15
	bclr	DelayTimeR, #15
	bra	__SPI1Interrupt_End
	
SPI_DoubleDelayTime_1:
	bset	Flags, #ResamplingMode
	bclr	Flags, #SingleInput
	bset	Flags, #HalfSampleRate
	bclr	BufferL_InOffset, #15
	bclr	BufferL_OutOffset, #15
	bclr	BufferL_OutOffset_New, #15
	bclr	DelayTimeL, #15
	bclr	BufferR_InOffset, #15
	bclr	BufferR_OutOffset, #15
	bclr	BufferR_OutOffset_New, #15
	bclr	DelayTimeR, #15
	bra	__SPI1Interrupt_End
	
SPI_DoubleDelayTime_2:
	bclr	Flags, #HalfSampleRate
	bset	Flags, #SingleInput
	bra	__SPI1Interrupt_End


; ---------------------------------------------------------------------------


SPI_EnableAudio:
	bset	Flags, #AudioEnabled
	
	bra	__SPI1Interrupt_End
	
	
; ---------------------------------------------------------------------------

	
__SPI1Interrupt_End:
	bclr	IFS0, #SPI1IF		; Clear the SPI interrupt flag
	pop.d	W2			; Restore W2, W3
	pop.d	W0			; Restore W0, W1
	pop	SR			; Restore SR flags
	retfie				; Return from ISR
	

; ---------------------------------------------------------------------------

	
; Call with delay time from UI PIC in W0
; Returns with actual delay time in ms/10 in W0
ConvertDelayTimeToMs:
; 0->49 = 0.0->4.9 (0.1 step)
; 50->74 = 5.0->9.8 (0.2 step)
; 75->114 = 10->49 (1 step)
; 115->189 = 50->198 (2 step)
; 190->219 = 200->345 (5 step)
; 220->252 = 350->670 (10 step)

	; <= 49 ?
	cp	N49
	bra	NN, ConvertDelayTimeToMs_1
	; <= 74 ?
	cp	N74
	bra	NN, ConvertDelayTimeToMs_2
	; <= 114 ?
	cp	N114
	bra	NN, ConvertDelayTimeToMs_3
	; <= 189 ?
	cp	N189
	bra	NN, ConvertDelayTimeToMs_4
	; <= 219 ?
	cp	N219
	bra	NN, ConvertDelayTimeToMs_5
	
; 220->252 = 350->670 (10 step)
	; W0 -= 185
	sub	#185, W0
	; W0 *= 10
	mul.uu	W0, #10, W0
	; W0 *= 10
	mul.uu	W0, #10, W0
	return
	
; 0->49 = 0.0->4.9 (0.1 step)
ConvertDelayTimeToMs_1:
	return

; 50->74 = 5.0->9.8 (0.2 step)
ConvertDelayTimeToMs_2:
	; W0 -= 25
	sub	#25, W0
	; W0 *= 2
	mul.uu	W0, #2, W0
	return

; 75->114 = 10->49 (1 step)
ConvertDelayTimeToMs_3:
	; W0 -= 65
	sub	#65, W0
	; W0 *= 10
	mul.uu	W0, #10, W0
	return

; 115->189 = 50->198 (2 step)
ConvertDelayTimeToMs_4:
	; W0 -= 90
	sub	#90, W0
	; W0 *= 20
	mul.uu	W0, #20, W0
	return

; 190->219 = 200->345 (5 step)
ConvertDelayTimeToMs_5:
	; W0 -= 150
	sub	#150, W0
	; W0 *= 10
	mul.uu	W0, #10, W0
	; W0 *= 5
	mul.uu	W0, #5, W0
	return
	

;===================================================================================

__DCIInterrupt:
	push.s				; Save C, Z, OV, N, DC flags & W0-W3 to shadow registers

; ---------------------------------------------------------------------------

; Fetch and invert L channel sample to W2
	mov	RXBUF2, W2
	neg	W2, W2
	
; Test for input clipping on L channel
	btsc	W2, #15			; Negative signal?
	bra	ClipLTestNeg
	mov	#0x6000, W0		; If W2 >= 0x6000 then signal clip
	cp	W2, W0
	bra	N, ClipLTestEnd
	bset	LATD, #10		; Turn on clip LED 1
	bset	T4CON, #TON		; Start Timer4
	bra	ClipLTestEnd
ClipLTestNeg:
	mov	#0xA001, W0		; If W2 <= 0xA000 then signal clip
	cp	W2, W0
	bra	NN, ClipLTestEnd
	bset	LATD, #10		; Turn on clip LED 1
	bset	T4CON, #TON		; Start Timer4
ClipLTestEnd:

; ---------------------------------------------------------------------------
	
; Scale (and clip if necessary) L channel sample to W1
	btsc	W2, #15			; Negative signal?
	bra	ScaleLNeg
	
	mov	#36700, W1		; Scale output = input x 1.12 (to give overall gain of 1)
	mul.su	W2, W1, W0
	rlc	W0, W0
	rlc	W1, W1
	
	btsc	W1, #15			; If now negative then overflow occurred
	mov	#0x7FFF, W1		; Cap W1 at 0x7FFF
	bra	ScaleLEnd
	
ScaleLNeg:
	mov	#36700, W1		; Scale output = input x 1.12 (to give overall gain of 1)
	mul.su	W2, W1, W0
	rlc	W0, W0
	rlc	W1, W1
	
	btss	W1, #15			; If now positive then overflow occurred
	mov	#0x8000, W1		; Cap W1 at 0x8000
ScaleLEnd:

	; Save input sample to InSampleL
	mov	W1,InSampleL
	
; ---------------------------------------------------------------------------

; Fetch and invert R channel sample to W2
	mov	RXBUF0, W2
	neg	W2, W2
	
; Test for input clipping on R channel
	btsc	W2, #15			; Negative signal?
	bra	ClipRTestNeg
	mov	#0x6000, W0		; If W2 >= 0x6000 then signal clip
	cp	W2, W0
	bra	N, ClipRTestEnd
	bset	LATD, #11		; Turn on clip LED 2
	bset	T5CON, #TON		; Start Timer5
	bra	ClipRTestEnd
ClipRTestNeg:
	mov	#0xA001, W0		; If W2 <= 0xA000 then signal clip
	cp	W2, W0
	bra	NN, ClipRTestEnd
	bset	LATD, #11		; Turn on clip LED 2
	bset	T5CON, #TON		; Start Timer5
ClipRTestEnd:

; ---------------------------------------------------------------------------
	
; Scale (and clip if necessary) R channel sample to W1
	btsc	W2, #15			; Negative signal?
	bra	ScaleRNeg
	
	mov	#36700, W1		; Scale output = input x 1.12 (to give overall gain of 1)
	mul.su	W2, W1, W0
	rlc	W0, W0
	rlc	W1, W1
	
	btsc	W1, #15			; If now negative then overflow occurred
	mov	#0x7FFF, W1		; Cap W1 at 0x7FFF
	bra	ScaleREnd
	
ScaleRNeg:
	mov	#36700, W1		; Scale output = input x 1.12 (to give overall gain of 1)
	mul.su	W2, W1, W0
	rlc	W0, W0
	rlc	W1, W1
	
	btss	W1, #15			; If now positive then overflow occurred
	mov	#0x8000, W1		; Cap W1 at 0x8000
ScaleREnd:

	; Save input sample to InSampleR
	mov	W1,InSampleR
	
; ---------------------------------------------------------------------------
	
; L channel processing ...
	mov	InSampleL, W1
	
	; If in single input mode add in R channel digital feedback
	btss	Flags, #SingleInput
	bra	DualInputMode

	; Add proportion of wet signal to input sample
	; W1 + OutSampleR * FeedbackFactorR => W1
	mov	OutSampleR, W2
	mov	FeedbackFactorR, W3
	mul.su	W2, W3, W2
	add	W1, W3, W1

	; Overflow?
	bra	NOV, DualInputMode
	
	bset	LATD, #10		; Turn on clip LED 1
	bset	T4CON, #TON		; Start Timer4

	; Negative?
	btss	SR, #N
	mov	#0x8000, W1

	; Positive?
	btsc	SR, #N
	mov	#0x7FFF, W1

DualInputMode:
	; Toggle OddSampleL flag
	btg	Flags, #OddSampleL

	; Add proportion of wet signal to input sample
	; W1 + OutSampleL * FeedbackFactorL => W1
	mov	OutSampleL, W2
	mov	FeedbackFactorL, W3
	mul.su	W2, W3, W2
	add	W1, W3, W1

	; Overflow?
	bra	NOV, NoOverflowL
	
	bset	LATD, #10		; Turn on clip LED 1
	bset	T4CON, #TON		; Start Timer4

	; Negative?
	btss	SR, #N
	mov	#0x8000, W1

	; Positive?
	btsc	SR, #N
	mov	#0x7FFF, W1
	
NoOverflowL:
	btss	Flags, #HalfSampleRate	; If running at full sample rate, skip next bit
	bra	FullSampleRateL

	btsc	Flags, #OddSampleL	; Store 1st of pair of samples in FIR_SampleInL1 (when OddSampleL flag set)
	mov	W1, FIR_SampleInL1
	btss	Flags, #OddSampleL	; Store 2nd of pair of samples in FIR_SampleInL2 (when OddSampleL flag clear)
	mov	W1, FIR_SampleInL2
	
	; If OddSampleL flag set, branch to resampling mode
	btsc	Flags, #OddSampleL
	bra	ResamplingModeL
	
	; Apply FIR filter
	bset	Flags, #DoFIR_L		; Flag that calculation needs to be made (in main loop)
	mov	FIR_ResultL, W1		; Get last result
	
	; Increment input pointer
	inc	BufferL_InOffset
	bclr	BufferL_InOffset, #15
	
	; Write input sample to SRAM
	mov	W1, W_SRAMData
	mov	BufferL_InOffset, W_SRAMAddress
	call	WriteSRAM
	
	bra	ReadBufferL
	
FullSampleRateL:
	; Increment input pointer
	inc	BufferL_InOffset
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	BufferL_InOffset, #15
	
	; Write input sample to SRAM
	mov	W1, W_SRAMData
	mov	BufferL_InOffset, W_SRAMAddress
	call	WriteSRAM
	
ReadBufferL:
	; If Buffer step != 0x100, branch to resampling mode
	mov	BufferL_OutStep, W0
	cp	N0x100
	bra	NZ, ResamplingModeL
	
	; Read output sample from SRAM
	mov	BufferL_OutOffset, W_SRAMAddress
	call	ReadSRAM
	mov	W_SRAMData, W1
	call	WriteOutL		; Write W1 to CODEC buffer
	
	; Increment output pointer
	inc	BufferL_OutOffset
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	BufferL_OutOffset, #15
	
	; If in resampling mode then skip next bit
	btsc	Flags, #ResamplingMode
	bra	ProcessingEndL
	
	; Increment BufferL_OutOffset_New
	inc	BufferL_OutOffset_New
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	BufferL_OutOffset, #15
	
	; Compare levels of old and new buffer location
	mov	W_SRAMData, W1
	mov	BufferL_OutOffset_New, W_SRAMAddress
	call	ReadSRAM
	sub	W_SRAMData, W1, W0
	btsc	SR, #N
	neg	W0, W0
	mov	#LEVEL_MATCH_THRESHOLD, W1
	cp	W1, W0
	bra	N, ProcessingEndL
	
	; Levels are near match so change buffer out pointer
	mov	BufferL_OutOffset_New, W0
	mov	W0, BufferL_OutOffset

	bra	ProcessingEndL
	
	
ResamplingModeL:
	; Read adjacent samples from SRAM -> W1 & W_SRAMData
	mov	BufferL_OutOffset, W_SRAMAddress
	call	ReadSRAM
	mov	W_SRAMData, W1
	inc	W_SRAMAddress, W_SRAMAddress
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	W_SRAMAddress, #15
	call	ReadSRAM
	
	; Interpolate between samples according to value of BufferL_OutOffset_L -> W1
	sub	W_SRAMData, W1, W0
	mov	BufferL_OutOffset_L, W2
	mul.su	W0, W2, W2
	lsr	W2, #8, W2
	and	#0b11111111, W3
	swap	W3
	ior	W3, W2, W2
	add	W1, W2, W1
	
	call	WriteOutL		; Write W1 to CODEC buffer
	
	; Increment output pointer by BufferL_OutStep
	mov	BufferL_OutStep, W0
	; If HalfSampleRate flag set then BufferL_OutStep >> 1
	btsc	Flags, #HalfSampleRate
	lsr	W0, #1, W0
	add	BufferL_OutOffset_L, WREG
	mov.b	WREG, BufferL_OutOffset_L
	lsr	W0, #8, W0
	add	BufferL_OutOffset
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	BufferL_OutOffset, #15
	
	
	; BufferL_OutStep < 0x100 ?
	mov	#0x100, W0
	cp	BufferL_OutStep
	bra	N, BufferL_OutStep_LT0x100
	
	
	; If BufferL_OutStep >= 0x100 ...
	; If (BufferL_InOffset - BufferL_OutOffset) < (DelayTimeL + BufferL_OutStep/0x100), then need to exit resampling mode
	
	; BufferL_InOffset - BufferL_OutOffset => W0
	mov	BufferL_OutOffset, W0
	sub	BufferL_InOffset, WREG
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	W0, #15
	
	; BufferL_OutStep / 0x100 => W1
	mov	BufferL_OutStep, W1
	lsr	W1, #8, W1
	
	; DelayTimeL + W1 => W1
	mov	DelayTimeL, W2
	add	W2, W1, W1
	
	; If W0 < W1, then need to exit resampling mode
	cp	W0, W1
	bra	LTU, ExitResamplingModeL
	bra	ProcessingEndL
	
	
BufferL_OutStep_LT0x100:
	; If BufferL_OutStep < 0x100 ...
	; If (BufferL_InOffset - BufferL_OutOffset) > (DelayTimeL - 1), then need to exit resampling mode
	
	; BufferL_InOffset - BufferL_OutOffset => W0
	mov	BufferL_OutOffset, W0
	sub	BufferL_InOffset, WREG
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	W0, #15
	
	; DelayTimeL - 1 => W1
	mov	DelayTimeL, W1
	dec	W1, W1
	
	; If W0 > W1, then need to exit resampling mode
	cp	W0, W1
	bra	GTU, ExitResamplingModeL
	bra	ProcessingEndL
	
	
ExitResamplingModeL:
	; Skip next bit if HalfSampleRate and OddSampleL flags set
	btss	Flags, #HalfSampleRate
	bra	$+6
	btsc	Flags, #OddSampleL
	bra	ProcessingEndL
	
	; BufferL_InOffset - DelayTimeL => BufferL_OutOffset
	mov	DelayTimeL, W0
	sub	BufferL_InOffset, WREG
	mov	W0, BufferL_OutOffset
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	BufferL_OutOffset, #15
	; BufferL_OutOffset_L = 0
	clr	BufferL_OutOffset_L
	; BufferL_OutStep = 0x100
	mov	#0x100, W0
	mov	W0, BufferL_OutStep
	
	bra	ProcessingEndL


; Writes W1 to CODEC buffer and handles dry mixing
WriteOutL:
	mov	W1, OutSampleL	; Save wet signal for digital feedback
	
	; Output = InSampleL * DryMixFactorL / 0x10000
	;          + W1      * WetMixFactorL / 0x10000
	
	mov	InSampleL, W2
	mov	DryMixFactorL, W3
	mul.su	W2, W3, W2
	
	mov	WetMixFactorL, W0
	mul.su	W1, W0, W0
	
	add	W3, W1, W1
	
.if (MUTE_AUDIO_DURING_STARTUP == 1)
	btss	Flags, #AudioEnabled
	clr	W1
.endif
	mov	W1, TXBUF2	; Write to CODEC buffer
	return
	

ProcessingEndL:
; ---------------------------------------------------------------------------
	
; R channel processing ...
	mov	InSampleR, W1

	; Toggle OddSampleR flag
	btg	Flags, #OddSampleR

	; Add proportion of wet signal to input sample
	; W1 + OutSampleR * FeedbackFactorR => W1
	mov	OutSampleR, W2
	mov	FeedbackFactorR, W3
	mul.su	W2, W3, W2
	add	W1, W3, W1

	; Overflow?
	bra	NOV, NoOverflowR
	
	bset	LATD, #11		; Turn on clip LED 2
	bset	T5CON, #TON		; Start Timer5

	; Negative?
	btss	SR, #N
	mov	#0x8000, W1

	; Positive?
	btsc	SR, #N
	mov	#0x7FFF, W1
	
NoOverflowR:
	btss	Flags, #HalfSampleRate	; If running at full sample rate, skip next bit
	bra	FullSampleRateR

	btsc	Flags, #OddSampleR	; Store 1st of pair of samples in FIR_SampleInR1 (when OddSampleR flag set)
	mov	W1, FIR_SampleInR1
	btss	Flags, #OddSampleR	; Store 2nd of pair of samples in FIR_SampleInR2 (when OddSampleR flag clear)
	mov	W1, FIR_SampleInR2
	
	; If OddSampleR flag set, branch to resampling mode
	btsc	Flags, #OddSampleR
	bra	ResamplingModeR
	
	; Apply FIR filter
	bset	Flags, #DoFIR_R		; Flag that calculation needs to be made (in main loop)
	mov	FIR_ResultR, W1		; Get last result
	
	; Increment input pointer
	inc	BufferR_InOffset
	bclr	BufferR_InOffset, #15
	
	; Write input sample to SRAM
	mov	W1, W_SRAMData
	mov	BufferR_InOffset, W_SRAMAddress
	bset	W_SRAMAddress, #15
	call	WriteSRAM
	
	bra	ReadBufferR
	
FullSampleRateR:
	; Increment input pointer
	inc	BufferR_InOffset
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	BufferR_InOffset, #15
	
	; Write input sample to SRAM
	mov	W1, W_SRAMData
	mov	BufferR_InOffset, W_SRAMAddress
	bset	W_SRAMAddress, #15
	btss	Flags, #SingleInput	; Don't write to SRAM if in single input mode
	call	WriteSRAM
	
ReadBufferR:
	; If Buffer step != 0x100, branch to resampling mode
	mov	BufferR_OutStep, W0
	cp	N0x100
	bra	NZ, ResamplingModeR
	
	; Read output sample from SRAM
	mov	BufferR_OutOffset, W_SRAMAddress
	btss	Flags, #SingleInput	; Don't set MSbit if in single input mode
	bset	W_SRAMAddress, #15
	call	ReadSRAM
	mov	W_SRAMData, W1
 	call	WriteOutR		; Write W1 to CODEC buffer
	
	; Increment output pointer
	inc	BufferR_OutOffset
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	BufferR_OutOffset, #15
	
	; If in resampling mode then skip next bit
	btsc	Flags, #ResamplingMode
	bra	ProcessingEndR
	
	; Increment BufferR_OutOffset_New
	inc	BufferR_OutOffset_New
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	BufferR_OutOffset, #15
	
	; Compare levels of old and new buffer location
	mov	W_SRAMData, W1
	mov	BufferR_OutOffset_New, W_SRAMAddress
	btss	Flags, #SingleInput	; Don't set MSbit if in single input mode
	bset	W_SRAMAddress, #15
	call	ReadSRAM
	sub	W_SRAMData, W1, W0
	btsc	SR, #N
	neg	W0, W0
	mov	#LEVEL_MATCH_THRESHOLD, W1
	cp	W1, W0
	bra	N, ProcessingEndR
	
	; Levels are near match so change buffer out pointer
	mov	BufferR_OutOffset_New, W0
	mov	W0, BufferR_OutOffset
	
	bra	ProcessingEndR
	
	
ResamplingModeR:
	; Read adjacent samples from SRAM -> W1 & W_SRAMData
	mov	BufferR_OutOffset, W_SRAMAddress
	btss	Flags, #SingleInput	; Don't set MSbit if in single input mode
	bset	W_SRAMAddress, #15
	call	ReadSRAM
	mov	W_SRAMData, W1
	inc	W_SRAMAddress, W_SRAMAddress
	btss	Flags, #SingleInput	; Don't set MSbit if in single input mode
	bset	W_SRAMAddress, #15
	call	ReadSRAM
	
	; Interpolate between samples according to value of BufferR_OutOffset_L -> W1
	sub	W_SRAMData, W1, W0
	mov	BufferR_OutOffset_L, W2
	mul.su	W0, W2, W2
	lsr	W2, #8, W2
	and	#0b11111111, W3
	swap	W3
	ior	W3, W2, W2
	add	W1, W2, W1
	
 	call	WriteOutR		; Write W1 to CODEC buffer
	
	; Increment output pointer by BufferR_OutStep
	mov	BufferR_OutStep, W0
	; If HalfSampleRate flag set then BufferR_OutStep >> 1
	btsc	Flags, #HalfSampleRate
	lsr	W0, #1, W0
	add	BufferR_OutOffset_L, WREG
	mov.b	WREG, BufferR_OutOffset_L
	lsr	W0, #8, W0
	add	BufferR_OutOffset
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	BufferR_OutOffset, #15
	
	
	; BufferR_OutStep < 0x100 ?
	mov	#0x100, W0
	cp	BufferR_OutStep
	bra	N, BufferR_OutStep_LT0x100
	
	
	; If BufferR_OutStep >= 0x100 ...
	; If (BufferR_InOffset - BufferR_OutOffset) < (DelayTimeR + BufferR_OutStep/0x100), then need to exit resampling mode
	
	; BufferR_InOffset - BufferR_OutOffset => W0
	mov	BufferR_OutOffset, W0
	sub	BufferR_InOffset, WREG
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	W0, #15
	
	; BufferR_OutStep / 0x100 => W1
	mov	BufferR_OutStep, W1
	lsr	W1, #8, W1
	
	; DelayTimeR + W1 => W1
	mov	DelayTimeR, W2
	add	W2, W1, W1
	
	; If W0 < W1, then need to exit resampling mode
	cp	W0, W1
	bra	LTU, ExitResamplingModeR
	bra	ProcessingEndR
	
	
BufferR_OutStep_LT0x100:
	; If BufferR_OutStep < 0x100 ...
	; If (BufferR_InOffset - BufferR_OutOffset) > (DelayTimeR - 1), then need to exit resampling mode
	
	; BufferR_InOffset - BufferR_OutOffset => W0
	mov	BufferR_OutOffset, W0
	sub	BufferR_InOffset, WREG
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	W0, #15
	
	; DelayTimeR - 1 => W1
	mov	DelayTimeR, W1
	dec	W1, W1
	
	; If W0 > W1, then need to exit resampling mode
	cp	W0, W1
	bra	GTU, ExitResamplingModeR
	bra	ProcessingEndR
	
	
ExitResamplingModeR:
	; Skip next bit if HalfSampleRate and OddSampleR flags set
	btss	Flags, #HalfSampleRate
	bra	$+6
	btsc	Flags, #OddSampleR
	bra	ProcessingEndR
	
	; BufferR_InOffset - DelayTimeR => BufferR_OutOffset
	mov	DelayTimeR, W0
	sub	BufferR_InOffset, WREG
	mov	W0, BufferR_OutOffset
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	BufferR_OutOffset, #15
	; BufferR_OutOffset_L = 0
	clr	BufferR_OutOffset_L
	; BufferR_OutStep = 0x100
	mov	#0x100, W0
	mov	W0, BufferR_OutStep
	
	bra	ProcessingEndR


; Writes W1 to CODEC buffer and handles dry mixing
WriteOutR:
	mov	W1, OutSampleR	; Save wet signal for digital feedback
	
	; Output = InSampleR * DryMixFactorR / 0x10000
	;          + W1      * WetMixFactorR / 0x10000
	
	mov	InSampleR, W2
	btsc	Flags, #SingleInput	; If in single input mode, take dry signal from L channel
	mov	InSampleL, W2
	mov	DryMixFactorR, W3
	mul.su	W2, W3, W2
	
	mov	WetMixFactorR, W0
	mul.su	W1, W0, W0
	
	add	W3, W1, W1
	
.if (MUTE_AUDIO_DURING_STARTUP == 1)
	btss	Flags, #AudioEnabled
	clr	W1
.endif
	mov	W1, TXBUF0	; Write to CODEC buffer
	return


ProcessingEndR:
; ---------------------------------------------------------------------------
	
; LFO processing ...
	
LFO_L:
	bclr	Flags, #SecondHalf
	mov	LFO_AngleL, W0
	
	; Saw Up ?
	mov	LFO_Wave, W1
	cp	W1, #2
	bra	Z, LFO_L_SawUp
	
	; Saw Down ?
	cp	W1, #3
	bra	Z, LFO_L_SawDown
	
	; Square ?
	cp	W1, #4
	bra	Z, LFO_L_Square
	
	btsc	W0, #15
	bset	Flags, #SecondHalf
	bclr	W0, #15
	btst.z	W0, #14
	bclr	W0, #14
	bra	Z, LFO_L_1
	
	; If 2nd and 4th quadrants, 16383 - W0 => W0
	mov	#16383, W1
	sub	W1, W0, W0
	
LFO_L_1:
	; Triangle ?
	cp0	LFO_Wave
	bra	NZ, LFO_L_Triangle
	
; Sine
	; Do sine table lookup
	bclr	W0, #0		; Clear odd bit to align with SineTable words
	mov	#psvoffset(SineTable), W1
	add	W0, W1, W0
	mov	[W0], W1
	
	; If 2nd half then invert value
	btsc	Flags, #SecondHalf
	neg	W1, W1
	
	bra	LFO_ScaleValueToDelayL
	
	
LFO_L_Triangle:
	; W0 * 2 => W1
	add	W0, W0, W1
	
	; If 2nd half then invert value
	btsc	Flags, #SecondHalf
	neg	W1, W1
	
	bra	LFO_ScaleValueToDelayL
	
	
LFO_L_SawUp:
	; Angle (W0) - 0x8000 => W1
	mov	#0x8000, W1
	sub	W0, W1, W1
	
	bra	LFO_ScaleValueToDelayL
	
	
LFO_L_SawDown:
	; 0x7FFF - Angle (W0) => W1
	mov	#0x7FFF, W1
	sub	W1, W0, W1
	
	bra	LFO_ScaleValueToDelayL
	
	
LFO_L_Square:
	; if Angle (W0) < 80000 then W1 = 0x8001
	; if Angle (W0) >= 80000 then W1 = 0x7FFF
	mov	#0x8001, W1
	btsc	W0, #15
	mov	#0x7FFF, W1
	
	bra	LFO_ScaleValueToDelayL


; W1 contains signed LFO value to be scaled by depth and applied to L delay time
LFO_ScaleValueToDelayL:
	; W1 (signed) * LFO_DepthL / 0x8000 => W3
	mov	LFO_DepthL, W2
	mul.su	W1, W2, W2
	rlc	W2, W2
	rlc	W3, W3
	
	; BaseDelayTimeL + W3 => W0
	mov	BaseDelayTimeL, W0
	add	W0, W3, W0
	
	; If negative set to 0
	btsc	SR, #N
	clr	W0
	
	; If overflow (i.e. >15bits) set to 0x7FFF
	btsc	SR, #OV
	mov	#0x7FFF, W0
	
	; Write to DelayTimeL and re-calculate BufferStepL
	mov	W0, DelayTimeL
	; If single input mode, DelayTimeL *= 2
	btsc	Flags, #SingleInput
	sl	DelayTimeL
	
	
; ---------------------------------------------------------------------------
	
	
; Convert DelayTimeL to BufferStepL

	btsc	Flags, #ResamplingMode		; Resampling mode?
	bra	DelayTimeL_To_BufferStepL_RM


; Non-resampling mode
	; BufferL_OutOffset_New = BufferL_InOffset - DelayTimeL
	mov	DelayTimeL, W0
	sub	BufferL_InOffset, WREG
	mov	W0, BufferL_OutOffset_New
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	BufferL_OutOffset_New, #15
	bra	DelayTimeL_To_BufferStepL_End


; Resampling mode
DelayTimeL_To_BufferStepL_RM:
	; BufferL_OutStep = 0x100 * (BufferL_InOffset - BufferL_OutOffset) / DelayTimeL
	
	; DelayTimeL == 0 ?
	cp0	DelayTimeL
	bra	Z, DelayTimeL_Zero
	
	; BufferL_InOffset - BufferL_OutOffset => W0
	mov	BufferL_OutOffset, W0
	sub	BufferL_InOffset, WREG
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	W0, #15
	; W0 * 0x100 => W1,W0
	mov	#0x100, W2
	mul.uu	W0, W2, W0
	; W1,W0 / DelayTimeL => W0
	mov	DelayTimeL, W2
	repeat	#17
	div.ud	W0, W2
	; If divide overflowed, set step to N_MAX_BUFFER_STEP
	bra	OV, DelayTimeL_MAX_BUFFER_STEP
	
	; If N_MAX_BUFFER_STEP < W0, set to N_MAX_BUFFER_STEP
	cp	N_MAX_BUFFER_STEP
	bra	LTU, DelayTimeL_MAX_BUFFER_STEP
	
	; If N_MIN_BUFFER_STEP > W0, set to N_MIN_BUFFER_STEP
	cp	N_MIN_BUFFER_STEP
	bra	GTU, DelayTimeL_MIN_BUFFER_STEP
	
	; W0 => BufferL_OutStep
	mov	W0, BufferL_OutStep
	bra	DelayTimeL_To_BufferStepL_End
	
DelayTimeL_MAX_BUFFER_STEP:
	; N_MAX_BUFFER_STEP => BufferL_OutStep
	mov	N_MAX_BUFFER_STEP, W0
	mov	W0, BufferL_OutStep
	bra	DelayTimeL_To_BufferStepL_End
	
DelayTimeL_MIN_BUFFER_STEP:
	; N_MIN_BUFFER_STEP => BufferL_OutStep
	mov	N_MIN_BUFFER_STEP, W0
	mov	W0, BufferL_OutStep
	bra	DelayTimeL_To_BufferStepL_End
	
; DelayTimeL == 0
DelayTimeL_Zero:
	; if (BufferL_InOffset - BufferL_OutOffset) < N_MAX_BUFFER_STEP
	; set BufferL_OutOffset = BufferL_InOffset
	
	; BufferL_InOffset - BufferL_OutOffset + 1 => W0
	dec	BufferL_OutOffset, WREG
	sub	BufferL_InOffset, WREG
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	W0, #15
	
	; N_MAX_BUFFER_STEP < W0 ?
	cp	N_MAX_BUFFER_STEP
	bra	LTU, DelayTimeL_MAX_BUFFER_STEP
	
	; BufferL_OutOffset = BufferL_InOffset
	mov	BufferL_InOffset, W0
	mov	W0, BufferL_OutOffset
	; BufferL_OutOffset_L = 0
	clr	BufferL_OutOffset_L
	; BufferL_OutStep = 0x100
	mov	#0x100, W0
	mov	W0, BufferL_OutStep
	bra	DelayTimeL_To_BufferStepL_End
	
	
DelayTimeL_To_BufferStepL_End:
; ---------------------------------------------------------------------------
	
LFO_R:
	bclr	Flags, #SecondHalf
	mov	LFO_AngleL, W0
	add	LFO_AngleROffset, WREG
	
	; Saw Up ?
	mov	LFO_Wave, W1
	cp	W1, #2
	bra	Z, LFO_R_SawUp
	
	; Saw Down ?
	cp	W1, #3
	bra	Z, LFO_R_SawDown
	
	; Square ?
	cp	W1, #4
	bra	Z, LFO_R_Square
	
	btsc	W0, #15
	bset	Flags, #SecondHalf
	bclr	W0, #15
	btst.z	W0, #14
	bclr	W0, #14
	bra	Z, LFO_R_1
	
	; If 2nd and 4th quadrants, 16383 - W0 => W0
	mov	#16383, W1
	sub	W1, W0, W0
	
LFO_R_1:
	; Triangle ?
	cp0	LFO_Wave
	bra	NZ, LFO_R_Triangle
	
; Sine
	; Do sine table lookup
	bclr	W0, #0		; Clear odd bit to align with SineTable words
	mov	#psvoffset(SineTable), W1
	add	W0, W1, W0
	mov	[W0], W1
	
	; If 2nd half then invert value
	btsc	Flags, #SecondHalf
	neg	W1, W1
	
	bra	LFO_ScaleValueToDelayR
	
	
LFO_R_Triangle:
	; W0 * 2 => W1
	add	W0, W0, W1
	
	; If 2nd half then invert value
	btsc	Flags, #SecondHalf
	neg	W1, W1
	
	bra	LFO_ScaleValueToDelayR
	
	
LFO_R_SawUp:
	; Angle (W0) - 0x8000 => W1
	mov	#0x8000, W1
	sub	W0, W1, W1
	
	bra	LFO_ScaleValueToDelayR
	
	
LFO_R_SawDown:
	; 0x7FFF - Angle (W0) => W1
	mov	#0x7FFF, W1
	sub	W1, W0, W1
	
	bra	LFO_ScaleValueToDelayR
	
	
LFO_R_Square:
	; if Angle (W0) < 80000 then W1 = 0x8001
	; if Angle (W0) >= 80000 then W1 = 0x7FFF
	mov	#0x8001, W1
	btsc	W0, #15
	mov	#0x7FFF, W1
	
	bra	LFO_ScaleValueToDelayR


; W1 contains signed LFO value to be scaled by depth and applied to R delay time
LFO_ScaleValueToDelayR:
	; W1 (signed) * LFO_DepthR / 0x8000 => W3
	mov	LFO_DepthR, W2
	mul.su	W1, W2, W2
	rlc	W2, W2
	rlc	W3, W3
	
	; BaseDelayTimeR + W3 => W0
	mov	BaseDelayTimeR, W0
	add	W0, W3, W0
	
	; If negative set to 0
	btsc	SR, #N
	clr	W0
	
	; If overflow (i.e. >15bits) set to 0x7FFF
	btsc	SR, #OV
	mov	#0x7FFF, W0
	
	; Write to DelayTimeR and re-calculate BufferStepR
	mov	W0, DelayTimeR
	; If single input mode, DelayTimeR *= 2
	btsc	Flags, #SingleInput
	sl	DelayTimeR


; ---------------------------------------------------------------------------


; Convert DelayTimeR to BufferStepR

	btsc	Flags, #ResamplingMode		; Resampling mode?
	bra	DelayTimeR_To_BufferStepR_RM


; Non-resampling mode
	; BufferR_OutOffset_New = BufferR_InOffset - DelayTimeR
	mov	DelayTimeR, W0
	sub	BufferR_InOffset, WREG
	mov	W0, BufferR_OutOffset_New
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	BufferR_OutOffset_New, #15
	bra	DelayTimeR_To_BufferStepR_End


; Resampling mode
DelayTimeR_To_BufferStepR_RM:
	; BufferR_OutStep = 0x100 * (BufferR_InOffset - BufferR_OutOffset) / DelayTimeR
	
	; DelayTimeR == 0 ?
	cp0	DelayTimeR
	bra	Z, DelayTimeR_Zero
	
	; BufferR_InOffset - BufferR_OutOffset => W0
	mov	BufferR_OutOffset, W0
	sub	BufferR_InOffset, WREG
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	W0, #15
	; W0 * 0x100 => W1,W0
	mov	#0x100, W2
	mul.uu	W0, W2, W0
	; W1,W0 / DelayTimeR => W0
	mov	DelayTimeR, W2
	repeat	#17
	div.ud	W0, W2
	; If divide overflowed, set step to N_MAX_BUFFER_STEP
	bra	OV, DelayTimeR_MAX_BUFFER_STEP
	
	; If N_MAX_BUFFER_STEP < W0, set to N_MAX_BUFFER_STEP
	cp	N_MAX_BUFFER_STEP
	bra	LTU, DelayTimeR_MAX_BUFFER_STEP
	
	; If N_MIN_BUFFER_STEP > W0, set to N_MIN_BUFFER_STEP
	cp	N_MIN_BUFFER_STEP
	bra	GTU, DelayTimeR_MIN_BUFFER_STEP
	
	; W0 => BufferR_OutStep
	mov	W0, BufferR_OutStep
	bra	DelayTimeR_To_BufferStepR_End
	
DelayTimeR_MAX_BUFFER_STEP:
	; N_MAX_BUFFER_STEP => BufferR_OutStep
	mov	N_MAX_BUFFER_STEP, W0
	mov	W0, BufferR_OutStep
	bra	DelayTimeR_To_BufferStepR_End
	
DelayTimeR_MIN_BUFFER_STEP:
	; N_MIN_BUFFER_STEP => BufferR_OutStep
	mov	N_MIN_BUFFER_STEP, W0
	mov	W0, BufferR_OutStep
	bra	DelayTimeR_To_BufferStepR_End
	
; DelayTimeR == 0
DelayTimeR_Zero:
	; if (BufferR_InOffset - BufferR_OutOffset) < N_MAX_BUFFER_STEP
	; set BufferR_OutOffset = BufferR_InOffset
	
	; BufferR_InOffset - BufferR_OutOffset + 1 => W0
	dec	BufferR_OutOffset, WREG
	sub	BufferR_InOffset, WREG
	btss	Flags, #SingleInput	; Don't clear MSbit if in single input mode
	bclr	W0, #15
	
	; N_MAX_BUFFER_STEP < W0 ?
	cp	N_MAX_BUFFER_STEP
	bra	LTU, DelayTimeR_MAX_BUFFER_STEP
	
	; BufferR_OutOffset = BufferR_InOffset
	mov	BufferR_InOffset, W0
	mov	W0, BufferR_OutOffset
	; BufferR_OutOffset_L = 0
	clr	BufferR_OutOffset_L
	; BufferR_OutStep = 0x100
	mov	#0x100, W0
	mov	W0, BufferR_OutStep
	bra	DelayTimeR_To_BufferStepR_End
	

DelayTimeR_To_BufferStepR_End:
; ---------------------------------------------------------------------------
	
	
; Step LFO_AngleL,LFO_AngleL_L by LFO_Step

	mov	LFO_Step, W0
	add	LFO_AngleL_L, WREG
	mov.b	WREG, LFO_AngleL_L
	lsr	W0, #8, W0
	add	LFO_AngleL


; ---------------------------------------------------------------------------
	
	clr	TXBUF3			; LS-8 bits of Left channel all 0
	clr	TXBUF1			; LS-8 bits of Right channel all 0
	
	; Save and clear idle counter
	mov	IdleCounter, W0
	mov	W0, IdleCounter_Prev
	clr	IdleCounter
	
	bclr	IFS2, #DCIIF		; Clear the DCI interrupt flag
	pop.s				; Restore C, Z, OV, N, DC flags & W0-W3
	retfie				; Return from ISR
	
;===================================================================================

	
;===================================================================================
	
	.end				; End of code in this file
